@@ -30,6 +30,7 @@
 #include "core/optimizer/graph_transformer_level.h"
 #include "core/optimizer/graph_transformer_mgr.h"
 #include "core/optimizer/insert_cast_transformer.h"
+#include "core/session/ep_graph_assignment_info.h"
 #include <mutex>
 #ifdef ENABLE_LANGUAGE_INTEROP_OPS
 #include "core/language_interop_ops/language_interop_ops.h"
@@ -58,6 +59,8 @@ class IExecutionProvider;
 class IOBinding;
 struct Notification;
 
+void reset_saturation_count();
+
 #ifdef ENABLE_TRAINING
 struct PartialGraphExecutionState;
 using OrtValueCache = InlinedHashMap<std::string, OrtValue>;
@@ -463,6 +466,34 @@ class InferenceSession {
    */
   std::pair<common::Status, const OutputDefList*> GetModelOutputs() const;
 
+  enum class SessionInputOutputType : uint8_t {
+    kInput = 0,
+    kOutput = 1,
+    kOverridableInitializer = 2
+  };
+
+  /**
+   * Get the OrtMemoryInfo for the inputs or outputs of the model.
+   *
+   * This is required for a user to know the location of the input/output when autoep selection is enabled.
+   */
+  common::Status GetInputOutputMemoryInfo(SessionInputOutputType type,
+                                          InlinedVector<const OrtMemoryInfo*>& memory_info) const;
+  /**
+   * Get the OrtEpDevice (if available) for the inputs of the model.
+   *
+   * This is required for a user to know the location of the input/output when autoep selection is enabled.
+   */
+  common::Status GetEpDeviceForInputs(InlinedVector<const OrtEpDevice*>& memory_info) const;
+
+  /**
+   * Get the OrtEpDevice (if available) for the outputs of the model.
+   *
+   * This is required for a user to validate that outputs will be placed on the expected device
+   * for external resource sharing.
+   */
+  common::Status GetEpDeviceForOutputs(InlinedVector<const OrtEpDevice*>& memory_info) const;
+
   /**
    * Get the current number of in-progress concurrent Run calls.
    */
@@ -474,6 +505,19 @@ class InferenceSession {
    */
   const std::vector<std::string>& GetRegisteredProviderTypes() const;
 
+  /**
+   * Get the registered Execution Providers.
+   *
+   * This method can be called after EP registration but before Initialize() completes.
+   * Used only for early validation of compiled model compatibility where accessing
+   * EPs through session state is not yet possible.
+   *
+   * @return const reference to the ExecutionProviders collection.
+   */
+  const ExecutionProviders& GetExecutionProviders() const noexcept {
+    return execution_providers_;
+  }
+
   /*
    * Get the options this session was initialized with.
    */
@@ -604,6 +648,40 @@ class InferenceSession {
   const Model& GetModel() const;
   const Environment& GetEnvironment() const;
 
+  void SetWeightDataType(const std::string& type) {
+    weight_data_type_ = type;
+  }
+
+  const std::string& GetWeightDataType() const {
+    return weight_data_type_;
+  }
+
+  void SetGraphHash(const std::string& hash) {
+    graph_hash_ = hash;
+  }
+
+  const std::string& GetGraphHash() const {
+    return graph_hash_;
+  }
+
+  void SetWeightHash(const std::string& hash) {
+    weight_hash_ = hash;
+  }
+
+  const std::string& GetWeightHash() const {
+    return weight_hash_;
+  }
+
+  uint32_t GetCurrentSessionId() const {
+    return session_id_;
+  }
+
+#if !defined(ORT_MINIMAL_BUILD)
+  const std::vector<const OrtEpAssignedSubgraph*>& GetEpGraphAssignmentInfo() const {
+    return this->ep_graph_assignment_info_;
+  }
+#endif  // !defined(ORT_MINIMAL_BUILD)
+
  protected:
 #if !defined(ORT_MINIMAL_BUILD)
 
@@ -676,6 +754,17 @@ class InferenceSession {
   // The file path of where the model was loaded. e.g. /tmp/test_squeezenet/model.onnx
   PathString model_location_;
 
+  // Input, Output and Weight tensor data types
+  std::string input_data_type_;
+  std::string output_data_type_;
+  std::string weight_data_type_;
+
+  // Graph hash of the model
+  std::string graph_hash_;
+
+  // Weight hash of the model
+  std::string weight_hash_;
+
  private:
   ORT_DISALLOW_COPY_ASSIGNMENT_AND_MOVE(InferenceSession);
   void SetLoggingManager(const SessionOptions& session_options,
@@ -881,15 +970,18 @@ class InferenceSession {
 
   struct Telemetry {
     Telemetry() : time_sent_last_() {}
-    uint32_t total_runs_since_last_ = 0;           // the total number of Run() calls since the last report
-    long long total_run_duration_since_last_ = 0;  // the total duration (us) of Run() calls since the last report
-    std::string event_name_;                       // where the model is loaded from: ["model_loading_uri", "model_loading_proto", "model_loading_istream"]
+    uint32_t total_runs_since_last_ = 0;                              // the total number of Run() calls since the last report
+    long long total_run_duration_since_last_ = 0;                     // the total duration (us) of Run() calls since the last report
+    std::string event_name_;                                          // where the model is loaded from: ["model_loading_uri", "model_loading_proto", "model_loading_istream"]
+    std::unordered_map<int64_t, long long> duration_per_batch_size_;  // the duration (us) of Run() calls per batch size since the last report
 
     TimePoint time_sent_last_;  // the TimePoint of the last report
     // Event Rate per provider < 20 peak events per second
     constexpr static long long kDurationBetweenSending = 1000 * 1000 * 60 * 10;  // duration in (us).  send a report every 10 mins
   } telemetry_;
 
+  mutable std::mutex telemetry_mutex_;  // to ensure thread-safe access to telemetry data
+
 #ifdef ONNXRUNTIME_ENABLE_INSTRUMENT
   bool session_activity_started_ = false;
   TraceLoggingActivity<telemetry_provider_handle> session_activity;
@@ -982,6 +1074,13 @@ class InferenceSession {
   // Enable nodestats collection
   std::optional<NodeStatsRecorder> node_stats_recorder_;
 #endif
+
+#if !defined(ORT_MINIMAL_BUILD)
+  // Information about the subgraphs/nodes assigned to each EP.
+  // A user gets this information via the OrtApi::GetEpGraphAssignmentInfo C API function.
+  std::vector<std::unique_ptr<OrtEpAssignedSubgraph>> ep_graph_assignment_info_storage_;
+  std::vector<const OrtEpAssignedSubgraph*> ep_graph_assignment_info_;
+#endif  // !defined(ORT_MINIMAL_BUILD)
 };
 
 struct SessionIOBinding {
